#!/usr/bin/env python
"""
$Id: resample_wrapper-FBF_1D.py,v 1.5 2004/03/03 19:34:30 maciek Exp $

Test wrapper for resampling stage, to be used with 1D sequences of 
spectra as stored in the FBF format. Thus, a typical input file would
have an FBF signature of .real8.1025 (where 1025 is the number of spectral
point samples).

Original Author: R.K.Garcia
Modified for 1D FBF input by: Maciek Smuga-Otto
"""

from mmap import *
from struct import *
from math import sin

from os import tempnam, fork, environ, execv, dup2, fdopen, pipe, waitpid, path
import sys
from time import sleep
import FBF
import Numeric

DEFAULT_TLWN = 15799.0

def get_infile_info( fbf_file ):
    return '%d,%d,%d' % (fbf_file.file.fileno(), fbf_file.grouping[0], fbf_file.record_size)

def createSHISfpref( tlwn = DEFAULT_TLWN ):
    """ there are only three entries for SHIS - and they vary only by detector
        0 - SW, 1 - MW, 2 - LW
        Spoofs an FBF object's hooks to conform to this wrapper's interface.
    """
    source_lwns = ( 15799.742, 15799.878, 15799.707 )
    spectrum_sizes = ( 4088, 4088, 2044 )
    QRF = 0 # Quality Reduction Factor
    DF = 0 # Decimation Factor
    
    fpref = file( tempnam(), 'wb+' )
    
    for slwn, size in zip( source_lwns, spectrum_sizes ):
        fpref.write( pack( 'ddLLL', tlwn, slwn, size, QRF, DF) )
    
    fpref.seek(0)
    return fpref

def fprefFromFBF( fbf_ref, spectrum_size, tlwn, qrf = 0, df = 0 ):
    """ given an FBF source laser wavenumber file, generate a reference database-
    format file. Assume no variability in target laser wavenumber (tlwn) or spectrum size"""
    source_lwns = fbf_ref.read(1,-1) # source laser wavenumbers
    fpref = file( tempnam(), 'wb+' )
    
    for slwn in source_lwns:
        # print ("%s %s %s" % (tlwn, slwn[0], spectrum_size)) #DEBUG
        fpref.write( pack( 'ddLLL', tlwn, slwn[0], spectrum_size, qrf, df) )
    
    fpref.seek(0)
    return fpref
    

# FIXME - two modes of operation here: one is to use a single entry from fpref (detector type),
# the other is to read from fpref in lockstep with input (GIFTS off-axis simulations)
# and possibly, some other mode!
        
def invoke( (fbf_out, fbf_in, fpref, reflen), exe ):
    rfd,wfd = pipe()
    writable0 = fdopen( wfd, 'w' )
    readable0 = fdopen( rfd, 'r' )
    rfd,wfd = pipe()
    writable1 = fdopen( wfd, 'w' )
    readable1 = fdopen( rfd, 'r' )
    
    pid = fork( )
    if (pid==0): # then I am the child
        nenv = { 'varInputSpectra': '%d,%d,%d' % \
                 (fbf_in.fp().fileno(), fbf_in.grouping[0], fbf_in.length()), 
                 'varOutputSpectra': '%d,%d,%d' % \
                 (fbf_out.fp().fileno(), fbf_out.grouping[0], fbf_out.length()),
                 'refSamplingSettings': '%d,%d' % (fpref.fileno(), reflen) }
        environ.update( nenv )
        writable0.close()
        readable1.close()
        print "child executing %s" % exe
        # change stdin and stdout
        dup2( readable0.fileno(), 0 ) #stdin
        dup2( writable1.fileno(), 1 ) #stdout
        execv( exe, [ exe ] )
        # end .. 
        
    readable0.close()
    writable1.close()
    return writable0, readable1, pid
    
def command( (writable0,readable1,pid), cmd ):
    print >>writable0, cmd
    writable0.flush()
    return readable1.readline()
    
def ping( link ):
    return command( link, "ping" )
        
def ident( link ):
    return command( link, "ident" )
        
def resample( link, out_idx, in_idx, ref_idx ):
    # print >> sys.stderr, 'resample %s %s %s' % (out_idx, in_idx, ref_idx) #DEBUG
    return command( link, 'resample %s %s %s' % (out_idx, in_idx, ref_idx) )

def doexit( link ):
    return command( link, 'exit' )
    

def main(f_in, f_out, f_ref = None, detector = None, tlwn = DEFAULT_TLWN):
    # input
    fbf_in = FBF.FBF( f_in )
    fbf_in.open() # open FBF file in read mode
    if fbf_in.array_type != 'd':
        print "ERROR: Expecting file to contain doubles, not single-precision floats or ints"
        sys.exit(-1)

    # reference
    if f_ref:
        fbf_ref = FBF.FBF(f_ref)
        spectrum_size = fbf_in.grouping[0]
        fref_ptr = fprefFromFBF( fbf_ref, spectrum_size, tlwn )
        reflen =  fbf_ref.length()
    else:
        fref_ptr = createSHISfpref( )
        reflen = 3

    # output
    out_dir, out_stem = path.split( f_out )
    fbf_out= FBF.FBF( stemname = out_stem, typename = 'real8',
                      grouping = fbf_in.grouping[:], dirname = out_dir )
    fbf_out.create()
    print "out shape: %s" % ([fbf_in.length(),] + fbf_out.grouping)
    fbf_out.block_write( 1, Numeric.zeros([fbf_in.length(),] + fbf_out.grouping, 'd') )
    fbf_out.fp().seek(0) # Dunno. starting to add stupid stuff that probably doesn't matter.
    # print "out length: %s" % fbf_out.length()
    

    print "forking executable stage.."
    wrp = invoke( (fbf_out, fbf_in, fref_ptr, reflen), './ResamplingStage' )
    print "ping result: %s" % ping( wrp ).strip()
    print "ident result: %s" % ident( wrp ).strip()
    
    if detector == -1: # GIFTS cube swath mode
        for idx in range( fbf_in.length() ):
            print "ping result: %s" % ping( wrp ).strip()
            print "resample result: %s" % resample( wrp, idx, idx, idx )
            # sleep(15) #DEBUG
    else: # SHIS mode
        for idx in range( fbf_in.length() ):
            print "resample result: %s" % resample( wrp, idx, idx, detector )
    print "exit request result: %s" % doexit( wrp )
    print "waiting for exit.."
    waitpid( wrp[2], 0 )

    fbf_in.close()
    fbf_out.close()
    print "done."

if __name__=='__main__':
    def usage():
        print "USAGE: %s (-s detector_index | -g fbf_ref) [-t tlwn] fbf_in fbf_out" % sys.argv[0]
        print "       where fbf_out is the [path_to/]stemname only,"
        print "       and exactly one of the two options is chosen:"
        print " -s: SHIS mode: detector_index is 0 for SW, 1 for MW and 2 for LW"
        print " -g: GIFTS simulation mode: fbf_ref is an FBF file of source laser wavenumbers"
        print " -t: explicitly set target laser wavenumber - defaults to 15799"
        sys.exit(1)

    import getopt # get command line arguments
    try:
        opts, args = getopt.getopt( sys.argv[1:], 's:g:t:' )
    except getopt.GetoptError:
        usage()
    if len(args) != 2:
        usage()

    mode = None
    shis_detector_index = -1 # GIFTS default
    f_ref = None
    tlwn = DEFAULT_TLWN
    
    for o, a in opts:
        if o == '-s':
            mode = 'SHIS'
            shis_detector_index = int(a)
        if o == '-g':
            mode = 'GIFTS'
            f_ref = a
        if o == '-t':
            tlwn = float(a)
    if not mode:
        usage()
        
    f_in, f_out = args
    main(f_in, f_out, f_ref, shis_detector_index, tlwn)
    sys.exit(0)
    